// This file is autogenerated. DO NOT EDIT
#pragma once
#include <robotpy_build.h>

namespace rpygen {

{% for name, tmpl_data in templates.items() %}
struct bind_{{ tmpl_data.x_qualname_ }}_{{ loop.index }} {
    bind_{{ tmpl_data.x_qualname_ }}_{{ loop.index }}(py::module &m, const char * clsName);
    void finish(const char *set_doc, const char *add_doc);
};
{% endfor %}

}; // namespace rpygen
