{#
    Pieces that go into an rpy-include file for a class

    - Trampoline base class (if applicable)
    - Template constructors/method fillers (if applicable)

#}

{% if cls.data.ignore or 
    (not cls.x_has_trampoline and not cls.data.template_params) %}
    {{ skip_generation() }}
{% endif %} 

// This file is autogenerated. DO NOT EDIT

#pragma once
#include <robotpy_build.h>

{% for inc in data.extra_includes_first %}
#include <{{ inc }}>
{% endfor %}

#include <{{ header.rel_fname }}>

{% for inc in data.extra_includes %}
#include <{{ inc }}>
{% endfor %}

{% if cls.x_has_trampoline %}
{% include 'cls_trampoline.hpp.j2' %}
{% endif %}

{% if cls.data.template_params %}
{% include 'cls_tmpl_impl.hpp.j2' %}
{% endif %}
