// This file is autogenerated. DO NOT EDIT
#include <robotpy_build.h>
{% if has_vcheck %}
#include <functional>
{% endif %}

{% for inc in data.extra_includes_first %}
#include <{{ inc }}>
{% endfor %}

#include <{{ header.rel_fname }}>

{% for inc in type_caster_includes %}
#include <{{ inc }}>
{% endfor %}

{% if x_has_operators %}
#include <pybind11/operators.h>
{% endif %}

{% for using in header.using.values() if using.using_type != "typealias" %}
using {{ using.raw_type }};
{% endfor %}

{% for cls in header.classes
   if not cls.data.ignore and cls.x_has_trampoline %}
#define RPYGEN_ENABLE_{{ cls.x_qualname_ }}_PROTECTED_CONSTRUCTORS
#include <rpygen/{{ cls.x_qualname_ }}.hpp>
{% endfor %}